# -*- coding: utf-8 -*-
#
#  box_cox.py
#  
#  Copyright 2022 Antoine Passemiers <antoine.passemiers@gmail.com>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import numpy as np
import torch
from sklearn.preprocessing import PowerTransformer

from portia.etel.var_transformation import VariableTransformation


class BoxCoxTransform(VariableTransformation):
    """Differentiable Box-Cox transform.

    Attributes:
        n_variables (int): Number of genes.
        bc_lambda_1 (:obj:`torch.nn.Parameter`): Vector of Box-Cox parameters, where i-th
            component corresponds to the parameter of the transform
            associated to gene i.
        _mu (:obj:`torch.nn.Parameter`): Mean vector, where i-th component is the average
            expression of gene i.
        _mu (:obj:`torch.nn.Parameter`): Standard deviation vector, where i-th component is the
            standard deviation of gene i.
    """

    def __init__(self, n_variables):
        VariableTransformation.__init__(self)
        self.n_variables = n_variables
        self.bc_lambda_1 = torch.nn.Parameter(torch.ones(1, self.n_variables))

        self._mu = None
        self._std = None
        self.sum_log_x = None

    def init(self, X_arr):
        """Initializes Box-Cox parameters.

        Each transform is fitted based on the marginal likelihood associated
        to one gene, instead of the joint likelihood.

        Args:
            X_arr (:obj:`np.ndarray`): NumPy array containing the input data.
        """
        transformer = PowerTransformer(method='box-cox', standardize=False)
        transformer.fit_transform(X_arr)
        self.bc_lambda_1.data = torch.FloatTensor(transformer.lambdas_.reshape(1, self.n_variables))
        self._mu = torch.FloatTensor(np.mean(X_arr, axis=0)).unsqueeze(0)
        self._std = torch.FloatTensor(np.std(X_arr, axis=0)).unsqueeze(0)
        self.sum_log_x = torch.sum(torch.log(torch.FloatTensor(X_arr)), dim=0)

    def _forward(self, X):
        Y1 = ((X ** self.bc_lambda_1) - 1.) / self.bc_lambda_1
        Y2 = torch.log(X)
        return torch.where(self.bc_lambda_1 != 0, Y1, Y2)

    def log_jacobian(self, X):
        return torch.sum((self.bc_lambda_1 - 1.) * self.sum_log_x)



# -*- coding: utf-8 -*-
#
#  variable.py
#
#  Copyright 2022 Antoine Passemiers <antoine.passemiers@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import numpy as np


class Variable:

    def __init__(self):
        self.parents = []
        self.bias = np.random.normal(0, 1)
        self.std = np.random.rand() * 2
        self.a = np.random.rand() * 3
        self.value = None

    def add_parent(self, var, coefficient):
        self.parents.append((var, coefficient))

    def sample(self):
        if self.value is None:
            value = 0
            for var, coefficient in self.parents:
                value += (1. / len(self.parents)) * coefficient * var.sample()
            value += np.random.normal(0, 1) * self.std
            self.value = self.a * np.exp(0.1 * value)
        return self.value

    def knock_out(self):
        self.value = 1e-5

    def reset(self):
        self.value = None



# -*- coding: utf-8 -*-
#
#  dataset.py
#  
#  Copyright 2022 Antoine Passemiers <antoine.passemiers@gmail.com>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import numpy as np


class Experiment:

    def __init__(self, _id, expression, knockout=None, knockdown=None, time=0):
        if knockdown is None:
            knockdown = []
        if knockout is None:
            knockout = []
        self._id = _id
        self.expression = expression
        self.knockout = knockout
        self.knockdown = knockdown
        self.time = time

    @property
    def id(self):
        return self._id

    @property
    def n_genes(self):
        return len(self.expression)


class GeneExpressionDataset:

    def __init__(self):
        self.experiments = dict()
        self.knockout = []
        self.knockdown = []
        self.X = []
        self.K = []

    def add(self, experiment):
        assert isinstance(experiment, Experiment)
        self.X.append(experiment.expression)
        k = np.zeros(len(experiment.expression), dtype=bool)
        k[experiment.knockout] = 1
        self.K.append(k)
        if self.n_genes > 0:
            assert self.n_genes == experiment.n_genes
        if experiment.id not in self.experiments:
            self.experiments[experiment.id] = list()
        if len(self.experiments[experiment.id]) > 0:
            assert(self.experiments[experiment.id][0].time <= experiment.time)
        self.experiments[experiment.id].append(experiment)
        
        # Add to knockout experiments
        gene_ids = tuple(experiment.knockout)
        if len(gene_ids) > 0:
            self.knockout.append(experiment)

        # Add to knockdown experiments
        gene_ids = tuple(experiment.knockdown)
        if len(gene_ids) > 0:
            self.knockdown.append(experiment)

    def compute_null_mutant_zscores(self, mean=None, std=None):

        # Knock-out experiments
        S = np.zeros((self.n_genes, self.n_genes))
        counts = np.zeros((self.n_genes, self.n_genes), dtype=np.int)

        # Compute statistics
        X = []
        for experiment in self.knockout:
            X.append(experiment.expression)
        X = np.asarray(X)
        if mean is None:
            if len(X) > 0:
                mean = np.mean(X, axis=0)
            else:
                mean = np.zeros(self.n_genes)
        if std is None:
            if len(X) > 0:
                std = np.std(X, axis=0) + 1e-50
            else:
                std = np.ones(self.n_genes)

        # Compute Z-scores
        for experiment in self.knockout:
            gene_ids = experiment.knockout
            if len(gene_ids) == 1:
                i = gene_ids[0]
                S[i, :] += np.abs(experiment.expression - mean) / std
                counts[i, :] += 1
        if np.sum(counts) > 0:
            mask = np.logical_or(counts == 0, counts.T == 0)
            S[~mask] = S[~mask] / counts[~mask]
            mean = np.mean(S[~mask])
            if mean == 0:
                S[~mask] = 1
            else:
                S[~mask] /= np.median(S[~mask])
            S[mask] = 1
            if np.sum(S) <= 0:
                S = np.ones((self.n_genes, self.n_genes))
        else:
            S = np.ones((self.n_genes, self.n_genes))

        np.fill_diagonal(S, 0)
        return S

    def asarray(self):
        return np.asarray(self.X)

    @property
    def n_samples(self):
        return len(self.X)

    @property
    def n_features(self):
        if len(self.X) == 0:
            return 0
        else:
            return len(self.X[0])

    @property
    def n_genes(self):
        return self.n_features

    def has_ko(self):
        return len(self.knockout) > 0



# -*- coding: utf-8 -*-
#
#  var_transformation.py
#  
#  Copyright 2022 Antoine Passemiers <antoine.passemiers@gmail.com>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

from abc import ABCMeta, abstractmethod

import torch


class VariableTransformation(torch.nn.Module, metaclass=ABCMeta):

    def __init__(self):
        torch.nn.Module.__init__(self)

    def forward(self, X):
        Y = self._forward(X)
        log_jac = self.log_jacobian(X)
        return Y, log_jac

    @abstractmethod
    def _forward(self, X):
        pass

    @abstractmethod
    def log_jacobian(self, X):
        pass



# -*- coding: utf-8 -*-
#
#  grn.py
#
#  Copyright 2022 Antoine Passemiers <antoine.passemiers@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import numpy as np


class GRN:

    def __init__(self, A, tf_idx):
        self.A = np.asarray(A, dtype=float)
        assert self.A.shape[0] == self.A.shape[1]
        self.tf_idx = np.asarray(tf_idx, dtype=int)

    def add_negatives_inplace(self):
        mask = np.any(self.A == 1, axis=1)
        self.A[mask, :] = np.nan_to_num(self.A[mask, :], nan=0)

    def as_symmetric(self):
        A = np.copy(self.A)

        # If one of the 2 directed edges is 1, then the undirected edge is 1
        mask = np.isnan(self.A)
        A[mask] = 0
        A = np.maximum(A, A.T)

        # If both directed edges are unknown, then the undirected edge is unknown
        mask = np.logical_and(mask, mask.T)
        A[mask] = np.nan

        # Create a new GRN from the adjacency matrix and TF list
        return GRN(A, self.tf_idx)

    def get_mask(self):
        mask = np.zeros(self.shape, dtype=bool)
        mask[self.tf_idx, :] = 1
        np.fill_diagonal(mask, 0)
        mask[np.isnan(self.A)] = 0
        return mask

    def binarize(self, n_top):

        # Missing values will be considered last (-inf is less than anything)
        # or won't be considered at all if n_top is not too big
        A = np.copy(self.A)
        np.nan_to_num(A, nan=-np.inf, copy=False)

        # Values associated to non-regulating genes will be considered last
        # or won't be considered at all if n_top is not too big
        mask = np.zeros(A.shape[0], dtype=bool)
        mask[self.tf_idx] = 1
        A[~mask, :] = -np.inf

        # Create an adjacency matrix from the `n_top` top values
        idx = np.argsort(A, axis=None)[-n_top:]
        idx = np.unravel_index(idx, self.shape)
        A = np.zeros(self.shape, dtype=np.uint8)
        A[idx[0], idx[1]] = 1

        # Create a new GRN from the adjacency matrix and TF list
        return GRN(A, self.tf_idx)

    @staticmethod
    def load_network(filepath, gene_names, tf_idx, from_string=False):
        n_genes = len(gene_names)
        if tf_idx is None:
            tf_idx = np.arange(n_genes)
        A = np.full((n_genes, n_genes), np.nan, dtype=float)
        gene_names = {gene_name: i for i, gene_name in enumerate(gene_names)}
        if from_string:
            lines = filepath.splitlines()
        else:
            with open(filepath, 'r') as f:
                lines = f.readlines()
        for line in lines:
            elements = line.rstrip().split()
            if len(elements) == 3:
                i = gene_names[elements[0]]
                j = gene_names[elements[1]]
                A[i, j] = float(elements[2])
            elif len(elements) == 2:
                i = gene_names[elements[0]]
                j = gene_names[elements[1]]
                A[i, j] = 1
        assert A.shape[0] == A.shape[1]
        np.fill_diagonal(A, 0)
        return GRN(A, tf_idx)

    @staticmethod
    def load_goldstandard(filepath, gene_names, tf_idx=None, from_string=False):
        return GRN.load_network(filepath, gene_names, tf_idx=tf_idx, from_string=from_string)

    def asarray(self):
        return self.A

    def shuffle(self):
        np.random.shuffle(self.A)

    def copy(self):
        A = np.copy(self.A)
        tf_idx = np.copy(self.tf_idx)
        return GRN(A, tf_idx)

    def __getitem__(self, key):
        return self.A[key]

    def __setitem__(self, key, value):
        self.A[key] = value

    def __len__(self):
        return self.n_genes

    @property
    def shape(self):
        return self.A.shape

    @property
    def n_genes(self):
        return self.A.shape[0]

    @property
    def n_edges(self):
        return int(np.nansum(self.A))

    @property
    def n_negatives(self):
        return np.sum(self.A == 0)
    



# -*- coding: utf-8 -*-
#
#  end_to_end.py
#  
#  Copyright 2022 Antoine Passemiers <antoine.passemiers@gmail.com>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import numpy as np
import torch

from portia.etel.box_cox import BoxCoxTransform
from portia.optim import LocalConvergence


def cov(X, ddof=0, aweights=None):
    """Differentiable covariance matrix estimation.

    Args:
        X (:obj:`torch.Tensor`): Matrix of observations, of shape (n_samples, n_genes).
        ddof (int, optional): When ddof=1, unbiased estimator is used.
        aweights (:obj:`torch.Tensor`): Array of sample weights, of shape (n_samples,).
            If None, all observations have equal weights.

    Returns:
        :obj:`torch.Tensor`: Estimated covariance matrix, of shape (n_genes, n_genes).
    """

    if (aweights is not None) and (not torch.is_tensor(aweights)):
        aweights = torch.DoubleTensor(aweights)

    if X.dim() == 1:
        X = X.view(-1, 1)

    if aweights is not None:
        weight_sum = torch.sum(aweights)
        mean = torch.sum(X * (aweights / weight_sum).unsqueeze(1), 0)
    else:
        mean = torch.mean(X, 0)

    if aweights is None:
        fact = X.shape[0] - ddof
    elif ddof == 0:
        fact = weight_sum
    elif aweights is None:
        fact = weight_sum - ddof
    else:
        sq_norm = torch.sum(aweights * aweights)
        fact = weight_sum - ddof * sq_norm / weight_sum

    centered = X.sub(mean.expand_as(X))

    if aweights is None:
        X_T = centered.t()
    else:
        X_T = torch.mm(torch.diag(aweights), centered).t()

    C = torch.mm(X_T, centered) / fact

    return C.squeeze()


def apply_optimal_transform(X, aweights=None, _lambda=0.8, max_n_iter=100, verbose=True):

    n_samples = X.shape[0]
    n_genes = X.shape[1]

    boxcox = BoxCoxTransform(n_genes)
    boxcox.init(X)
    X = torch.DoubleTensor(X)

    boxcox.bc_lambda_1.data = torch.clamp(boxcox.bc_lambda_1.data, -60, 60)

    optimizer = torch.optim.Adam([
        {'params': boxcox.parameters(), 'lr': 1e-2}
    ])
    convergence = LocalConvergence(5, max_n_iter=max_n_iter, tau=1e-7)

    if verbose:
        print('Starting gradient descent.')

    iteration = 1
    while not convergence():

        optimizer.zero_grad()

        X_prime, log_jac1 = boxcox(X)
        assert not np.isnan(X_prime.data.numpy()).any()

        C = cov(X_prime, aweights=aweights)
        C_reg = _lambda * torch.diag(torch.diagonal(C)) + (1. - _lambda) * C

        try:
            L = torch.linalg.cholesky(C_reg)
            Theta = torch.cholesky_inverse(L)
            logdet = -2. * torch.sum(torch.log(torch.diagonal(L)))
            # Theta = torch.inverse(C_reg)
            # logdet = torch.logdet(Theta)
        except RuntimeError:
            if verbose:
                print('Invalid value encountered. Falling back to sequential PORTIA.')
            boxcox.init(X)
            break

        ll = -0.5 * (n_samples - 1.) * torch.trace(torch.mm(Theta, C)) + 0.5 * n_samples * logdet
        # ll = -0.5 * n_samples * torch.trace(torch.mm(Theta, C)) + 0.5 * n_samples * logdet

        loss = -ll - log_jac1

        loss.backward()

        optimizer.step()
        boxcox.bc_lambda_1.data = torch.clamp(boxcox.bc_lambda_1.data, -60, 60)

        convergence.step(loss.item())

        if verbose:
            print(f'Loss at iteration {iteration}: {loss.item()}')
        iteration += 1

    boxcox.standardize = True
    X_prime, _ = boxcox(X)
    return X_prime.data.numpy()



# -*- coding: utf-8 -*-
#
#  local_convergence.py
#  
#  Copyright 2022 Antoine Passemiers <antoine.passemiers@gmail.com>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import numpy as np


class LocalConvergence:

    def __init__(self, n_steps_without_improvements, max_n_iter=1000, tau=1e-5):
        self.n_steps_without_improvements = n_steps_without_improvements
        self.max_n_iter = max_n_iter
        self.tau = tau
        self.iterations = 0
        self.nwi = 0
        self.best_loss = np.nan_to_num(np.inf)

    def step(self, loss):
        gain = (self.best_loss - loss) / np.abs(self.best_loss)
        if gain > self.tau:
            self.nwi = 0
        else:
            self.nwi += 1
        if loss < self.best_loss:
            self.best_loss = loss
        self.iterations += 1

    def __call__(self):
        if self.iterations >= self.max_n_iter:
            return True
        else:
            return self.nwi >= self.n_steps_without_improvements



# -*- coding: utf-8 -*-
#
#  la.py
#  
#  Copyright 2022 Antoine Passemiers <antoine.passemiers@gmail.com>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import numpy as np


def all_linear_regressions(X, _lambda=0.05):
    """Perform m linear regressions in cubic time,
    using the same covariance matrix.

    aiu2 = u^T @ T_j
         = u^T @ T - u^T @ aiun @ vai^T
         = (T^T @ u)^T - u^T @ aiun @ vai^T
         = aiu - u^T @ aiun @ vai^T
    vai2 = T_j @ v
         = T @ v - aiun @ vai^T @ v
         = vai - aiun @ vai^T @ v
    """
    n_genes = X.shape[1]
    S = _lambda * np.eye(n_genes) + (1. - _lambda) * np.cov(X.T)
    T = np.linalg.inv(S)
    B = T @ X.T @ X

    V = -S - np.eye(n_genes)
    VAI = np.dot(V.T, T)
    den1 = 1. + np.sum(V * T, axis=0)

    B -= T * (np.sum(np.dot(X, VAI) * X, axis=0) / den1)[np.newaxis, :]
    AIU2 = T - VAI * (np.diag(T) / den1)[np.newaxis, :]
    VAIV = np.sum(VAI * V, axis=0)
    XAIU2X = np.sum(np.dot(X, AIU2) * X, axis=0)

    B -= (VAI - T * (VAIV / den1)[np.newaxis, :]) * (XAIU2X / (1. + np.sum(V * AIU2, axis=0)))[np.newaxis, :]

    np.fill_diagonal(B, 0)
    return B


def all_linear_regressions_slow(X, _lambda=0.05):
    n_genes = X.shape[1]
    S = _lambda * np.eye(n_genes) + (1. - _lambda) * np.cov(X.T)
    T = np.linalg.inv(S)
    B = np.zeros((n_genes, n_genes))
    u = np.zeros(n_genes)
    for j in range(n_genes):

        # Remove i-th row of covariance matrix using Sherman-Morrison formula
        u[:] = 0
        u[j] = 1
        v = -S[:, j]
        v[j] = -(S[j, j] - 1)
        aiu = np.dot(T, u)
        vai = np.dot(v, T)
        T_j = T - np.outer(aiu, vai) / (1. + np.dot(v, aiu))

        # Remove i-th column of covariance matrix using Sherman-Morrison formula
        v[j] = 0
        aiu = np.dot(T_j.T, u)
        vai = np.dot(v, T_j.T)
        T_j = (T_j.T - np.outer(aiu, vai) / (1. + np.dot(v, aiu))).T

        # Remove j-th variable (explained variable) from the set of explanatory variables
        T_j[:, j] = 0
        T_j[j, :] = 0
        T_j[j, j] = 0

        # Compute OLS estimator (see Gauss-Markov theorem)
        B[:, j] = T_j.dot(X.T).dot(X[:, j])
    np.fill_diagonal(B, 0)
    return B


def all_linear_regressions_naive(X, _lambda=0.05):
    n_genes = X.shape[1]
    beta = []
    for j in range(n_genes):
        mask = np.ones(n_genes, dtype=bool)
        mask[j] = 0
        X_j = X[:, mask]
        y_j = X[:, j]
        cov = _lambda * np.eye(n_genes - 1) + (1. - _lambda) * np.cov(X_j.T)
        w = np.zeros(n_genes)
        w[mask] = np.linalg.inv(cov) @ X_j.T @ y_j
        assert not np.any(np.isnan(w))
        beta.append(w)
    beta = np.asarray(beta).T
    np.fill_diagonal(beta, 0)
    return beta


def partial_inv(mat, tf_idx):
    n_rows = mat.shape[0]
    _Theta = np.zeros((n_rows, n_rows), dtype=mat.dtype)
    _J = np.eye(n_rows)[:, tf_idx]
    _Theta[tf_idx, :] = np.linalg.solve(mat, _J).T
    return _Theta



# -*- coding: utf-8 -*-
#
#  core.py
#  
#  Copyright 2022 Antoine Passemiers <antoine.passemiers@gmail.com>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import enum
import collections.abc
import warnings

import scipy.spatial
from sklearn.preprocessing import PowerTransformer, StandardScaler

from portia.correction import apply_correction
from portia.dataset import GeneExpressionDataset
from portia.exceptions import *
from portia.la import *


class Method(enum.Enum):

    # Fast power transforms
    FAST = enum.auto()

    # End-to-end inference of the optimal parameters for the power transforms
    END_TO_END = enum.auto()

    # No power transform
    NO_TRANSFORM = enum.auto()

    @staticmethod
    def from_string(name):
        formatted_name = name.strip().lower().replace('-', ' ').replace('_', ' ')
        if formatted_name == 'fast':
            return Method.FAST
        elif formatted_name == 'end to end':
            return Method.END_TO_END
        elif formatted_name == 'no transform':
            return Method.NO_TRANSFORM
        else:
            raise InvalidParameterError(
                f'Unknown method name "{name}". '
                f'Method should be "fast", "end-to-end" or "no-transform".')


def run(dataset, tf_idx=None, method='fast', lambda1=0.8, lambda2=0.05, n_iter=1000,
        normalize=True, return_sign=False, verbose=True):
    """Infers a Gene Regulatory Network (GRN) from gene expression data.

    The output network is a score matrix of shape (n_genes, n_genes).

    Args:
        dataset (:obj:`portia.GeneExpressionDataset`): Gene expression dataset.
        tf_idx (iterable): List of non-redundant 0-based indices.
            Each index in `tf_idx` corresponds to a potential transcription factor,
            and should be in the range [0, n_genes - 1]. The list should contain at
            most n_genes elements. If None, then all genes are considered as potentially
            regulating genes.
        method (str): Non-linear transformation of gene expression data.
            Should be either "fast", "end-to-end" or "no-transform".
        lambda1 (float): Shrinkage parameter for the estimation of the covariance matrix.
            The higher, the smaller the condition number of the covariance matrix will be.
        lambda2 (float): Tiknonov regularization rate for inferring directional information
            from gene expression exclusively.
        n_iter (int): Maximum number of iterations ("end-to-end" method only).
        normalize (bool): Whether to normalize each experiment separately or not.
            Normalization is based on the median.
        return_sign (bool): Whether to return the signs of inferred regulatory links.
        verbose (bool): Whether to print information to the standard output.

    Returns:
        :obj:`np.ndarray`: A real-valued estimation of the adjacency matrix of the GRN.
            The scores range from 0 to 1.
        :obj:`np.ndarray`: Returned only when `return_sign` is True. An integer matrix
            of same shape as the score matrix, where 1 indicates a positive regulatory relation,
            -1 a negative regulatory relation, and 0 no relation at all.
    """
    
    # Arbitrarily-small value used for numerical stability purposes
    epsilon = 1e-50

    # Check parameters
    method = Method.from_string(method)
    if not isinstance(dataset, GeneExpressionDataset):
        raise InvalidParameterError('Parameter "dataset" should be of type `portia.GeneExpressionDataset`.')
    if tf_idx is None:
        tf_idx = np.arange(0, dataset.n_genes)
    if not isinstance(tf_idx, collections.abc.Iterable):
        raise InvalidParameterError('Parameter "tf_idx" should be an iterable (e.g. list or array), or None.')
    tf_idx = np.asarray(tf_idx)
    try:
        _min = np.min(tf_idx)
        _max = np.max(tf_idx)
        if (_min < 0) or (_max >= dataset.n_genes):
            raise InvalidParameterError('Parameter "tf_idx" should contain indices in [0, n_genes - 1].')
        if len(np.unique(tf_idx)) != tf_idx.shape[0]:
            raise InvalidParameterError('Parameter "tf_idx" should not contain redundant indices.')
    except:
        raise InvalidParameterError('Parameter "tf_idx" should contain 0-based indices.')
    if not (0 <= lambda1 <= 1):
        raise InvalidParameterError('Shrinkage parameter "lambda1" should be in [0, 1].')
    if not (0 <= lambda2 <= 1):
        raise InvalidParameterError('Shrinkage parameter "lambda2" should be in [0, 1].')
    if not isinstance(normalize, bool):
        raise InvalidParameterError('Parameter "normalize" should be of boolean type.')
    if not isinstance(verbose, bool):
        raise InvalidParameterError('Parameter "verbose" should be of boolean type.')

    # Get expression data
    _X = dataset.asarray()
    n_samples = _X.shape[0]
    n_genes = _X.shape[1]
    if verbose:
        print(f'Gene expression matrix of shape ({n_samples}, {n_genes})')

    # Check missing values
    if np.any(np.isnan(_X)):
        raise NotImplementedError('Missing values are not supported (yet).')

    # Check the presence of negative values
    mask = np.any(_X < 0, axis=0)
    if np.any(mask):
        warnings.warn(
            'Expression data contain negative values. '
            'It is recommended to provide raw data instead.')
        _X[:, mask] -= (np.min(_X[:, mask], axis=0) - 1e-2)

    # Normalize experiments
    if normalize:
        quantiles = np.quantile(_X, 0.5, axis=0)[np.newaxis, :]
        quantiles[quantiles == 0] = 1
        _X = _X / quantiles

    # Compute sample weights (reduce redundancy)
    if n_samples >= n_genes:
        _P = 1. - scipy.spatial.distance.cdist(_X, _X, metric='correlation')
        theta = 0.8
        counts = np.zeros(n_samples)
        idx = np.where(_P > theta)
        np.add.at(counts, idx[0], 1)
        np.add.at(counts, idx[1], 1)
        weights = 1.0 / (1. + np.asarray(counts, dtype=np.float))
    else:
        weights = np.ones(n_samples) / n_samples

    # Box-Cox transform
    mask = np.asarray([(len(np.unique(_X[:, i])) > 1) for i in range(n_genes)], dtype=np.bool_)
    if method == Method.FAST:
        if np.sum(mask) > 0:
            power_transform = PowerTransformer(method='box-cox', standardize=True)
            _X_transformed = _X
            _X_transformed[:, mask] = power_transform.fit_transform(_X[:, mask] + epsilon)
        else:
            _X_transformed = _X + epsilon
    elif method == Method.END_TO_END:
        if np.sum(mask) > 0:
            _X_transformed = _X
            from portia.end_to_end import apply_optimal_transform
            _X_transformed[:, mask] = apply_optimal_transform(
                    _X[:, mask] + 1e-45, aweights=weights, _lambda=lambda1,
                    max_n_iter=100, verbose=verbose)
        else:
            _X_transformed = _X
    else:
        _X_transformed = _X

    # Centering and scaling
    scaler = StandardScaler(copy=False, with_mean=True, with_std=True)
    _X_transformed = scaler.fit_transform(_X_transformed)

    # Compute empirical covariance matrix
    _S = np.cov(_X_transformed.T, ddof=1, aweights=weights)

    # Apply covariance matrix shrinkage
    _S_bar = lambda1 * np.eye(n_genes) + (1. - lambda1) * _S

    # Compute precision matrix
    if (tf_idx is not None) and (len(tf_idx) < 0.5 * n_genes):
        _Theta = partial_inv(_S_bar, tf_idx)
    else:
        _Theta = np.linalg.inv(_S_bar)

    # Gene bias correction
    _M = np.abs(_Theta)
    if tf_idx is not None:
        mask = np.zeros((n_genes, n_genes))
        mask[tf_idx, :] = 1
        np.fill_diagonal(mask, 0)
        _M *= mask
        _M[tf_idx, :] = apply_correction(_M[tf_idx, :], method='rcw')
    else:
        _M = apply_correction(_M, method='rcw')

    # No gene self-regulation
    np.fill_diagonal(_M, 0)

    # Break symmetry using linear regressions.
    # All OLS estimators are efficiently computed from a single
    # covariance matrix, using Sherman-Morrison formula.
    beta = all_linear_regressions(_X_transformed, _lambda=lambda2)
    beta = np.abs(beta)
    np.fill_diagonal(beta, 0)
    div = np.maximum(beta, beta.T)
    div_mask = (div > 0)
    beta[div_mask] /= div[div_mask]
    _M = _M * beta

    # Compute null-mutant Z-scores (if KOs are available)
    if dataset.has_ko():
        _F = dataset.compute_null_mutant_zscores()

        # Combine corrected precision matrix and Z-scores
        norm = np.linalg.norm(_F)
        if norm > 0:
            _F = _F / norm
        _M_bar = _M * _F ** 2
    else:
        _M_bar = _M

    # Promote scores from "hub" genes
    _M_bar *= np.std(_M_bar, axis=1)[:, np.newaxis]

    # Ensure that scores are in the range [0, 1]
    _range = _M_bar.max() - _M_bar.min()
    if _range > 0:
        _M_bar = (_M_bar - _M_bar.min()) / _range

    if return_sign:
        signs = np.sign(_Theta)
        np.fill_diagonal(_Theta, 0)
        return _M_bar, signs
    else:
        return _M_bar


def rank_scores(scores, gene_names, tf_names=None, limit=np.inf, diagonal=False):
    """Generator that iterates over top-scoring regulatory links by decreasing order of score.

    Args:
        scores (:obj:`np.ndarray`): A real-valued estimation of the adjacency matrix of the GRN.
            This can be, for example, the output of `portia.run` function.
        gene_names (iterable): List of gene names, in the same order as the corresponding 
            rows and columns in `scores`.
        tf_names (iterable, optional): List of regulating genes. Each gene name should necessarily
            be in `gene_names`. If None, then all genes are considered as potentially regulating genes.
        limit (int, optional): Maximum number of top-scoring links to generate.
        diagonal (bool, optional): Whether to include the diagonal of the score matrix.

    Yields:
        tuple: The next regulatory link to be reported. It is represented as a tuple containing
            three elements: the name of the regulating gene, the name of the regulated gene,
            and the corresponding score in the real-valued adjacency matrix.
    """

    if tf_names is None:
        tf_names = set(gene_names)

    idx = np.argsort(np.abs(scores.flatten(order='C')))[::-1]
    idx = np.unravel_index(idx, scores.shape, order='C')

    n_predictions = 0

    for k in range(len(idx[0])):
        i = idx[0][k]
        j = idx[1][k]

        if gene_names[i] in tf_names:

            # Discard scores on diagonal if not needed
            if (not diagonal) and (i == j):
                continue

            yield gene_names[i], gene_names[j], scores[i, j]
            n_predictions += 1
            if n_predictions >= limit:
                break



# -*- coding: utf-8 -*-
#
#  correction.py
#  
#  Copyright 2022 Antoine Passemiers <antoine.passemiers@gmail.com>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import numpy as np


def apply_correction(C, method='rcw', eps=1e-50, ignore_diag=True):
    is_square = (C.shape[0] == C.shape[1])
    C = np.copy(C)
    if is_square and ignore_diag:
        np.fill_diagonal(C, 0)
    if min(C.shape[0], C.shape[1]) < 2:
        return C
    else:
        if ignore_diag:
            n_nonzero_genes_i = np.sum(C != 0, axis=1)
            n_nonzero_genes_i = np.maximum(n_nonzero_genes_i, 1)
            n_nonzero_genes_j = np.sum(C != 0, axis=0)
            n_nonzero_genes_j = np.maximum(n_nonzero_genes_j, 1)
            Fi = C.sum(axis=1) / n_nonzero_genes_i
            Fj = C.sum(axis=0) / n_nonzero_genes_j
            F = C.sum() / (np.sum(n_nonzero_genes_i) + eps)
        else:
            Fi = np.mean(C, axis=1)
            Fj = np.mean(C, axis=0)
            F = np.mean(C)
        if method == 'apc':
            # Average product correction, as proposed by
            # S.D. Dunn et al.
            apc = Fi[:, np.newaxis] * Fj[np.newaxis, :]
            if F != 0:
                apc = apc / F
            C -= apc
        elif method == 'asc':
            # Average sum correction, as proposed by
            # S.D. Dunn et al.
            asc = Fi[:, np.newaxis] + Fj[np.newaxis, :] - F
            C -= asc
        else:
            # Row-column weighting
            rcw = Fi[:, np.newaxis] * Fj[np.newaxis, :]
            C /= (2. * (rcw + eps))
        return C



# -*- coding: utf-8 -*-
#
#  symmetry.py
#
#  Copyright 2022 Antoine Passemiers <antoine.passemiers@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import numpy as np

from portia.gt.grn import GRN


def matrix_symmetry(A):
    if isinstance(A, GRN):
        tf_idx = A.tf_idx
        A = A.asarray()
        A = A[tf_idx, :][:, tf_idx]

    A = np.asarray(A)
    Aa = 0.5 * (A - A.T)
    As = 0.5 * (A + A.T)
    mask = ~np.logical_or(np.isnan(Aa), np.isnan(As))
    norm_aa = np.linalg.norm(Aa[mask])
    norm_as = np.linalg.norm(As[mask])
    if norm_as + norm_aa == 0:
        return 1
    else:
        return (norm_as - norm_aa) / (norm_as + norm_aa)



# -*- coding: utf-8 -*-
#
#  utils.py
#
#  Copyright 2022 Antoine Passemiers <antoine.passemiers@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import numpy as np


def nan_to_min(x):
    x = np.copy(x)
    mask = np.isnan(x)
    if np.sum(~mask) == 0:
        x[mask] = 0
    else:
        x[mask] = np.min(x[~mask])
    return x



# -*- coding: utf-8 -*-
#
#  evaluation.py
#
#  Copyright 2022 Antoine Passemiers <antoine.passemiers@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import os
import sys

import numpy as np

from portia.gt.causal_structure import CausalStructure
from portia.gt.grn import GRN
from portia.gt.utils import nan_to_min
from portia.gt.topology import _evaluate, _all_connected


def graph_theoretic_evaluation(filepath, G_target, G_pred, tf_mask=None, method='top'):

    method = method.strip().lower()
    assert method in {'top', 'auto'}

    # Check TF mask
    if tf_mask is None:
        n_genes = len(G_target)
        tf_mask = np.ones(n_genes, dtype=bool)

    # Ensure GRNs
    if not isinstance(G_target, GRN):
        tf_idx = np.where(tf_mask)[0]
        G_target = GRN(G_target, tf_idx)
    if not isinstance(G_pred, GRN):
        tf_idx = np.where(tf_mask)[0]
        G_pred = GRN(G_pred, tf_idx)

    # Ensure there is at least one negative
    # If there is no negative, negatives are added by filling all
    # missing values with zeroes in rows for which there is experimental evidence.
    # Indeed, since regulatory relationships are mostly discovered by experiments
    # (e.g. KO), these experiments also provide evidence for the absence of regulatory
    # relationships for the remaining target genes.
    if G_target.n_negatives == 0:
        G_target.add_negatives_inplace()

    # Goldstandard adjacency matrix
    A = np.copy(G_target.asarray())

    # Goldstandard undirected (symmetric) adjacency matrix
    AU = G_target.as_symmetric().asarray()

    # Convert predicted scores to binary adjacency matrix
    if method == 'top':
        use_top = True
    else:
        use_top = (len(np.unique(G_target.asarray())) > 2)
    if use_top:
        n_edges = G_target.n_edges
        A_binary_pred = G_pred.binarize(n_edges).asarray()
        assert int(np.sum(A_binary_pred)) == n_edges
    else:
        A_binary_pred = (G_pred.asarray() >= np.max(G_pred.asarray()))
        n_edges = np.sum(A_binary_pred)

    # Fill missing values
    np.nan_to_num(A, nan=0, copy=False)
    np.nan_to_num(AU, nan=0, copy=False)
    A_binary_pred = nan_to_min(A_binary_pred)

    if (filepath is not None) and os.path.exists(filepath):
        data = np.load(filepath)
        C = data['C']
        CU = data['CU']
    else:
        # Find all pairs of connected vertices
        C = _all_connected(A, tf_mask)

        # Find all pairs of connected vertices (undirected edges)
        CU = _all_connected(AU, np.ones(AU.shape[0]))

        # If `i` indirectly regulates `j`, then an undirected regulatory relationship
        # exists between `j` and `i`, and vice versa
        # This step is necessary because we didn't compute node connectivity when tf_mask[i]
        # is False, leading to missing values in `CU`
        CU = np.maximum(CU, CU.T)

        # Save regulatory relationship matrices
        if filepath is not None:
            np.savez(filepath, C=C, CU=CU)
    assert np.all(CU >= C)

    # No self-regulation
    np.fill_diagonal(A, 0)
    np.fill_diagonal(C, 0)
    np.fill_diagonal(CU, 0)

    # Categorise predictions based on local causal structures
    results = {'T': _evaluate(A, A_binary_pred, C, CU, tf_mask)}

    # Convert the matrix of causal structures into a relevance matrix
    relevance = CausalStructure.array_to_relevance(results['T'])

    # Filter predictions and keep only meaningful ones:
    # no self-regulation, and no prediction for TFs for which there is no
    # experimental evidence, based on the goldstandard
    mask = G_target.get_mask()
    y_relevance = relevance[mask]
    y_pred = G_pred[mask]

    # Fill missing predictions with the lowest score.
    # For fair comparison of the different GRN inference methods,
    # the same number of predictions should be reported.
    # If a method reports less regulatory links than what is present
    # in the goldstandard, missing values will be consumed until
    # the right number of predictions (`n_edges`) is reached.
    y_pred = nan_to_min(y_pred)

    # Get the indices of the first `n_edges` predictions,
    # sorted by decreasing order of importance
    idx = np.argsort(y_pred)[-n_edges:][::-1]

    # Compute importance weights
    weights = 1. / np.log2(1. + np.arange(1, len(idx) + 1))

    # Discounted Cumulative Gain
    dcg = np.sum(weights * y_relevance[idx])

    # Ideal Discounted Cumulative Gain
    idcg = 4. * np.sum(weights)

    # Normalized Discounted Cumulative Gain (ranges between 0 and 1)
    ndcg = dcg / idcg

    # For safety reasons
    if n_edges == 0:
        ndcg = 0

    # Store NDCG score
    results['score'] = ndcg
    return results


def plot_fp_types(ax, G_target, G_pred, T, tf_mask=None, n_pred=300, remove_labels=True, legend=False):
    import matplotlib.pyplot as plt  # pylint: disable=import-error

    # Check TF mask
    if tf_mask is None:
        n_genes = len(G_target)
        tf_mask = np.ones(n_genes, dtype=bool)

    # Ensure GRNs
    n_genes = len(G_target)
    if not isinstance(G_target, GRN):
        tf_idx = np.arange(n_genes)
        G_target = (G_target, tf_idx)
    if not isinstance(G_pred, GRN):
        tf_idx = np.arange(n_genes)
        G_target = (G_pred, tf_idx)

    n_pred = int(n_pred)
    n_links = G_target.n_edges

    plt.rcParams['figure.figsize'] = [12, 4]

    A = G_target.asarray()
    S = G_pred.asarray()

    mask = G_target.get_mask()
    _as = A[mask]
    ys = np.nan_to_num(S[mask], nan=0)
    types = T[mask]

    idx = np.argsort(ys)[::-1][:n_pred]
    types = types[idx]
    ys = ys[idx]
    _as = _as[idx]
    assert not np.any(np.isnan(types))
    assert not np.any(np.isnan(ys))
    assert not np.any(np.isnan(_as))
    used_colors = set()
    for i in range(len(ys)):
        if not _as[i]:
            if types[i] > 0:
                if types[i] == CausalStructure.CHAIN:
                    color = 'green'
                elif types[i] == CausalStructure.FORK:
                    color = 'orange'
                elif types[i] == CausalStructure.CHAIN_REVERSED:
                    color = 'red'
                elif types[i] == CausalStructure.COLLIDER:
                    color = 'cyan'
                elif types[i] == CausalStructure.UNDIRECTED:
                    color = 'purple'
                elif types[i] == CausalStructure.SPURIOUS_CORRELATION:
                    color = 'black'
                elif types[i] == CausalStructure.TRUE_POSITIVE:
                    color = 'grey'
                else:
                    color = 'none'
                if color != 'none':
                    if legend and (color not in used_colors):
                        ax.bar(i, ys[i], width=0.4, color=color, label=CausalStructure.to_string(types[i]))
                        used_colors.add(color)
                    else:
                        ax.bar(i, ys[i], width=0.4, color=color)
        if i + 1 >= n_links:
            break
    # plt.axvline(x=n_links, linewidth=1, linestyle='--', color='gray', alpha=0.5)
    ax.set_yscale('log')
    if remove_labels:
        plt.tick_params(labelleft=False)
        ax.set(yticklabels=[" "])
        ax.axes.yaxis.set_ticklabels([" "])
        ax.axes.yaxis.set_visible(False)
    else:
        ax.set_ylabel('log(predicted score)', fontname='Century gothic')
        ax.set_xlabel('Top n predictions', fontname='Century gothic')

    if legend:
        plt.legend()



# -*- coding: utf-8 -*-
#
#  causal_structure.py
#
#  Copyright 2022 Antoine Passemiers <antoine.passemiers@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import enum

import numpy as np


class CausalStructure(enum.IntEnum):

    # True positive regulatory link
    TRUE_POSITIVE = enum.auto()

    # Indirectly regulated genes
    CHAIN = enum.auto()

    # D-connected genes
    FORK = enum.auto()

    # Indirectly regulated genes, with link predicted
    # in the wrong direction
    CHAIN_REVERSED = enum.auto()

    # Collider (d-separated genes)
    COLLIDER = enum.auto()

    # Undirected link (d-separated genes)
    UNDIRECTED = enum.auto()

    # Spurious correlation
    SPURIOUS_CORRELATION = enum.auto()

    # True negative
    TRUE_NEGATIVE = enum.auto()

    # False negative
    FALSE_NEGATIVE = enum.auto()

    @staticmethod
    def array_to_relevance(M):
        """Converts matrix of causal structures (enums) to relevance scores.

        Args:
            M: Matrix of causal structure of shape (n_genes, n_genes), where `[i, j]`
                is the local causal structure (enum) of gene pair `(i, j)`.

        Returns:
            Array `R` of same shape as `M`, where `R[i, j]` is the relevance of
            reporting the `(i, j)` gene pair as a regulatory link.
        """
        R = np.zeros(M.shape, dtype=np.uint8)

        # True positives
        R[M == CausalStructure.TRUE_POSITIVE] = 4

        # Forward chains
        R[M == CausalStructure.CHAIN] = 2

        # D-connected variables (that are not TPs nor forward chains)
        R[M == CausalStructure.FORK] = 1
        R[M == CausalStructure.CHAIN_REVERSED] = 1

        # D-separated variables (that are indirectly related)
        R[M == CausalStructure.COLLIDER] = 0.5
        R[M == CausalStructure.UNDIRECTED] = 0.5

        # Remaining cases (spurious correlations)
        R[M == CausalStructure.SPURIOUS_CORRELATION] = 0
        return R

    @staticmethod
    def to_string(i):
        """String representation of a causal structure (enum).

        Args:
            i (CausalStructure): enum representing a local causal structure.

        Returns:
            The string representation of the local causal structure.
        """
        if i == CausalStructure.TRUE_POSITIVE:
            return 'True positive'
        elif i == CausalStructure.CHAIN:
            return 'Chain'
        elif i == CausalStructure.FORK:
            return 'Fork'
        elif i == CausalStructure.CHAIN_REVERSED:
            return 'Chain (reversed)'
        elif i == CausalStructure.COLLIDER:
            return 'Collider'
        elif i == CausalStructure.UNDIRECTED:
            return 'Undirected'
        elif i == CausalStructure.SPURIOUS_CORRELATION:
            return 'Spurious correlation'
        elif i == CausalStructure.TRUE_NEGATIVE:
            return 'True negative'
        elif i == CausalStructure.FALSE_NEGATIVE:
            return 'False negative'
        else:
            return 'Unknown'



# -*- coding: utf-8 -*-
#
#  topology.py
#
#  Copyright 2022 Antoine Passemiers <antoine.passemiers@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import numpy as np

from portia.gt.causal_structure import CausalStructure


def _all_connected(A, tf_mask):
    """Naive algorithm for finding all pairs of connected vertices.

    The performance of this algorithm will strongly depend on the topology
    of the input graph, but it proved to be quite efficient in practice.

    Note: possible improvement, compute accessibility matrix (e.g. with Floyd-Warshall algorithm)
    """
    C_old = np.copy(np.asarray(A, dtype=np.uint8))
    while True:
        C_new = np.logical_or(C_old, np.dot(C_old, C_old))
        if np.all(C_new == C_old):
            break
        C_old = C_new
    return C_new


def _evaluate(A, A_pred, C, CU, tf_mask):
    A = np.asarray(A, dtype=bool)
    A_pred = np.asarray(A_pred, dtype=bool)
    C = np.asarray(C, dtype=bool)
    CU = np.asarray(CU, dtype=bool)
    T = np.zeros(A.shape, np.uint8)
    tf_mask = np.asarray(tf_mask, dtype=bool)
    n_genes = A.shape[0]

    assert A.shape == (n_genes, n_genes)
    assert A_pred.shape == (n_genes, n_genes)
    assert C.shape == (n_genes, n_genes)
    assert CU.shape == (n_genes, n_genes)
    assert tf_mask.shape == (n_genes,)

    mask = np.logical_and(A_pred, ~A)

    # Chains
    M = C
    T[np.logical_and(mask, M)] = int(CausalStructure.CHAIN)
    np.logical_and(mask, ~M, out=mask)

    # Chains (reversed)
    M = C.T
    T[np.logical_and(mask, M)] = int(CausalStructure.CHAIN_REVERSED)
    np.logical_and(mask, ~M, out=mask)

    # Forks
    M = np.dot(C.T, C).astype(bool)
    T[np.logical_and(mask, M)] = int(CausalStructure.FORK)
    np.logical_and(mask, ~M, out=mask)

    # Colliders
    M = np.dot(C, C.T).astype(bool)
    T[np.logical_and(mask, M)] = int(CausalStructure.COLLIDER)
    np.logical_and(mask, ~M, out=mask)

    # Undirected (d-separated)
    M = CU
    T[np.logical_and(mask, M)] = int(CausalStructure.UNDIRECTED)
    np.logical_and(mask, ~M, out=mask)

    # Spurious correlations
    T[mask] = int(CausalStructure.SPURIOUS_CORRELATION)

    # True negatives
    mask = np.logical_and(~A_pred, ~A)
    T[mask] = int(CausalStructure.TRUE_NEGATIVE)

    # False negatives
    mask = np.logical_and(~A_pred, A)
    T[mask] = int(CausalStructure.FALSE_NEGATIVE)

    # True positives
    mask = np.logical_and(A_pred, A)
    T[mask] = int(CausalStructure.TRUE_POSITIVE)

    # True negatives
    T[~tf_mask, :] = int(CausalStructure.TRUE_NEGATIVE)
    np.fill_diagonal(T, int(CausalStructure.TRUE_NEGATIVE))

    return T



# -*- coding: utf-8 -*-
# setup.py
# author: Antoine Passemiers

import re

from setuptools import setup


packages = [
    'portia',
    'portia.etel',
    'portia.gt',
    'portia.insilico',
    'portia.optim'
]

with open('README.md', 'r') as f:
    long_description = f.read()


def get_property(prop, project):
    # Solution proposed in:
    # https://stackoverflow.com/questions/17791481/creating-a-version-attribute-for-python-packages-without-getting-into-troubl
    result = re.search(r'{}\s*=\s*[\'"]([^\'"]*)[\'"]'.format(prop), open(project + '/__init__.py').read())
    return result.group(1)


setup(
    name='portia-grn',
    version=get_property('__version__', 'portia'),
    author='Antoine Passemiers',
    author_email='antoine.passemiers@kuleuven.be',
    packages=packages,
    url='https://github.com/AntoinePassemiers/PORTIA',
    description='PORTIA: Fast and Accurate Inference of Gene Regulatory Networks through Robust Precision Matrix Estimation',
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
    install_requires=[
        'numpy>=1.19.5',
        'scipy',
        'scikit-learn'
    ],
)



# -*- coding: utf-8 -*-
#
#  dataset.py
#
#  Copyright 2022 Antoine Passemiers <antoine.passemiers@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import numpy as np

from portia.insilico.variable import Variable


def make_dataset(n, n_genes, n_tfs, ko=0, mko=0, sparsity=0.997):
    """Generate a GRN and gene expression dataset.

    Args:
        n (int): Number of gene expression measurements.
        n_genes (int): Total number of genes (GRN size).
        n_tfs (int): Number of putative transcripion factors.
            Should be lower than the number of genes.
        ko (int, optional): Number of single-gene KO experiments.
            The sum of `ko` and `mko` should not exceed `n`.
        mko (int, optional): Number of multiple-gene KO experiments.
            The sum of `ko` and `mko` should not exceed `n`.
        sparsity (float, optional): The percentage of gene pairs that
            do not form a regulatory relationship.

    Returns:
        X (np.ndarray): Gene expression matrix, of shape (n, n_genes).
        Z (np.ndarray): Binary matrix of shape (n, n_genes), indicating which genes
            have been knocked out in each experiment.
        A (np.ndarray): Goldstandard GRN, a binary matrix of shape (n, n_genes).
        tf_mask (np.ndarray): A binary vector indicating which genes are putative
            transcription factors.
    """
    assert n_tfs <= n_genes
    assert 0 < sparsity < 1
    assert 0 <= ko
    assert 0 <= mko
    assert ko + mko <= n
    A = (np.random.rand(n_genes, n_genes) < 1. - sparsity)
    np.fill_diagonal(A, 0)
    A[n_tfs:, :] = 0
    A[:n_tfs, :n_tfs] = np.triu(A[:n_tfs, :n_tfs])

    genes = [Variable() for _ in range(n_genes)]
    for i in range(n_tfs):
        for j in range(n_genes):
            if A[i, j]:
                coefficient = np.random.normal(0, 1)
                genes[j].add_parent(genes[i], coefficient)

    X = []
    Z = []
    for i in range(n):

        z = np.zeros(n_genes, dtype=bool)

        if ko > 0:
            j = np.random.randint(0, n_tfs)
            genes[j].knock_out()
            z[j] = 1
            ko -= 1
        elif mko > 0:
            idx = np.random.randint(0, n_tfs, size=10)
            for j in idx:
                genes[j].knock_out()
                z[j] = 1
            mko -= 1
        Z.append(z)

        X.append([gene.sample() for gene in genes])
        for gene in genes:
            gene.reset()
    X = np.nan_to_num(X)
    Z = np.asarray(Z, dtype=bool)
    tf_mask = np.zeros(n_genes, dtype=bool)
    tf_mask[:n_tfs] = True

    assert np.all(X >= 0)

    return X, Z, A, tf_mask
